[data-md-color-scheme="slate"] {
  --md-default-bg-color: #120512;
  --md-primary-fg-color: #06d49e;
}

[data-md-color-scheme="default"] {
  --md-primary-fg-color: #06d49e;
}

.code-output {
    font-size: 0.5rem; /* Adjust the size as needed */
}
